-- Paragraph and text classes

ebook.par_classes = {}
ebook.par_class_tags = {}
ebook.par_ids = {}
ebook.text_classes = {}
ebook.text_class_tags = {}

local par_class_attr = luatexbase.attributes["parclass"]
ebook.debug(">>> par_class_attr = " .. par_class_attr)

local par_id_attr = luatexbase.attributes["parid"]
ebook.debug(">>> par_id_attr = " .. par_id_attr)

local text_class_attr = luatexbase.attributes["textclass"]
ebook.debug(">>> text_class_attr = " .. text_class_attr)

local as_pic_attr = luatexbase.attributes["aspic"]
ebook.debug(">>> as_pic_attr = " .. as_pic_attr)

function ebook.define_par_class(id, class, tag)
	ebook.debug(">>> Par class #" .. id .. " class=" .. class .. " tag=" .. tag)
	ebook.par_classes[id] = class
	ebook.par_class_tags[id] = tag
end

-- Normalize XML identifier
-- ucwmac uses ':' as a hierarchy separator, we translate it to '-'
function ebook.normalize_id(id)
	id = string.gsub(id, ":", "-")
	return id
end

function ebook.define_par_id(id, name)
	ebook.debug(">>> Par id #" .. id .. " name=" .. name)
	ebook.par_ids[id] = ebook.normalize_id(name)
end

function ebook.define_text_class(id, class, tag)
	if tag == "" then
		tag = "span"
	end
	ebook.debug(">>> Text class #" .. id .. " class=" .. class .. " tag=" .. tag)
	ebook.text_classes[id] = class
	ebook.text_class_tags[id] = tag
end

-- Raw HTML nodes

local raw_html_whatsit = luatexbase.new_whatsit("raw_html")
ebook.debug(">>> raw_html_whatsit = " .. raw_html_whatsit)

local anchor_whatsit = luatexbase.new_whatsit("html_anchor")
ebook.debug(">>> anchor_whatsit = " .. anchor_whatsit)

-- class is a text_class
-- class==nil means "keep class active in the moment of definition"
-- class<0 means "do not switch class on this node"
local function new_raw_html(html, class)
	local x = node.new(8, 8)	-- user-defined whatsit
	x.user_id = raw_html_whatsit
	x.type = 115			-- string
	x.value = html
	if class then
		if class < 0 then
			node.unset_attribute(x, text_class_attr)
		else
			node.set_attribute(x, text_class_attr, class)
		end
	end
	return x
end

local function new_anchor(id)
	local x = node.new(8, 8)	-- user-defined whatsit
	x.user_id = anchor_whatsit
	x.type = 115			-- string
	x.value = ebook.normalize_id(id)
	return x
end

function ebook.raw_html(html, class)
	local x = new_raw_html(html, class)
	node.write(x)
end

function ebook.html_anchor(id)
	local x = new_anchor(id)
	node.write(x)
end

local function new_fixme()
	return new_raw_html("<b>FIXME</b>")
end

local function html_escape(s)
	s = string.gsub(s, "&", "&amp;")
	s = string.gsub(s, "<", "&lt;")
	s = string.gsub(s, ">", "&gt;")
	return s
end

-- List operations

local function list_append(head, tail, new)
	if tail then
		tail.next = new
	else
		head = new
	end
	return head, node.slide(new)
end

local function list_append_html(head, tail, html, class)
	return list_append(head, tail, new_raw_html(html, class or -1))
end

local function new_vfill()
	local n = node.new(12, 0)	-- glue, userskip
	n.stretch = 65536
	n.stretch_order = 1
	return n
end

local function new_penalty(p)
	local n = node.new(14)		-- penalty
	n.penalty = p
	return n
end

-- Math classes and replacements

ebook.math_chars = {}
ebook.math_classes = {}
ebook.math_repls = {}

local math_repl_attr = luatexbase.attributes["mathrepl"]
ebook.debug(">>> math_repl_attr = " .. math_repl_attr)

local math_learn_whatsit = luatexbase.new_whatsit("math_learn")
ebook.debug(">>> math_learn_whatsit = " .. math_learn_whatsit)

function ebook.define_math_repl(id, note, html, class, type)
	ebook.debug(string.format(">>> Math repl #%d for=%s html=%s class=%d type=%d", id, note, html, class, type))
	ebook.math_repls[id] = { ["html"] = html, ["class"] = class, ["type"] = type }
end

local function math_char_def(fam, char, html, class)
	ebook.debug(">>> Math char: fam=" .. fam .. " char=" .. char .. " html=" .. html .. " class=" .. (class or -1))
	if not ebook.math_chars[fam] then
		ebook.math_chars[fam] = {}
		ebook.math_classes[fam] = {}
	end
	ebook.math_chars[fam][char] = html
	if class then
		ebook.math_classes[fam][char] = class
	end
end

function ebook.define_math_chars(fam, chars, class)
	for char in string.gmatch(chars, ".") do
		math_char_def(fam, string.byte(char), char, class)
	end
end

function ebook.define_math_char_map(fam, chars, class, codept)
	for char in string.gmatch(chars, ".") do
		math_char_def(fam, string.byte(char), unicode.utf8.char(codept), class)
		codept = codept + 1
	end
end

function ebook.define_math_code(fam, char, html, class)
	math_char_def(fam, char, html, class)
end

function ebook.add_math_learn(html, class)
	x = node.new(8, 8)	-- user-defined whatsit
	x.user_id = math_learn_whatsit
	x.type = 115		-- string
	x.value = html
	if class then
		node.set_attribute(x, text_class_attr, class)
	end
	node.write(x)
end

-- Generating output

local out
local out_counter = 0
out_name = "???"
local finished
local need_html_epilog

local function html_prolog()
	out:write("<?xml version='1.0' encoding='utf-8'?>\n")
	local lang = html_escape(ebook.get_config('meta', 'lang'))
	out:write(string.format("<html xmlns='http://www.w3.org/1999/xhtml' lang='%s' xml:lang='%s'>\n", lang, lang))
	out:write("<head>\n")
	out:write(string.format("\t<title>%s</title>\n", html_escape(ebook.get_config('meta', 'title'))))
	out:write("\t<link href='../style/default.css' rel='stylesheet' type='text/css' />\n")
	out:write("</head>\n")
	out:write("<body>\n\n")
	need_html_epilog = 1
end

local function html_epilog()
	out:write("\n")
	out:write("</body>\n")
	out:write("</html>\n")
end

local function close_output()
	if out then
		if need_html_epilog then
			html_epilog()
			need_html_epilog = nil
		end
		out:close()
		out = nil
	end
end

function ebook.new_output()
	close_output()
	out_counter = out_counter + 1
	ebook.out_name = string.format("part%03d.html", out_counter)
	local name = "epub/html/" .. ebook.out_name
	out = io.open(name, "w")
	if not out then
		tex.error("Cannot create " .. name)
	end
	html_prolog()
end

function ebook.special_output(name)
	close_output()
	ebook.out_name = name
	out = io.open(name, "w")
	if not out then
		tex.error("Cannot create " .. name)
	end
end

function ebook.bye_bye()
	close_output()
	finished = 1
end

-- Before we initialize the output explicitly, drop everything
ebook.special_output("/dev/null")

-- Conversion of TeX lists to HTML

local delayed_start_of_par
local delayed_space
local current_class = 0

local function enter_class(class)
	local tag = ebook.text_class_tags[class] or "span"
	local cls = ebook.text_classes[class] or ""
	if cls == "" then
		out:write("<" .. tag .. ">")
	else
		out:write("<" .. tag .. " class='" .. cls .. "'>")
	end
	current_class = class
end

local function leave_class()
	if current_class ~= 0 then
		local tag = ebook.text_class_tags[current_class] or "span"
		out:write("</" .. tag .. ">")
		current_class = 0
	end
end

local special_chars = {
	["<"] = "&lt;",
	[">"] = "&gt;",
	["&"] = "&amp;",
}

local lig_table = {
	["--"] = "–",
	["---"] = "—",
}

local function add_html(html, class)
	if delayed_start_of_par then
		out:write(delayed_start_of_par)
		delayed_start_of_par = nil
	end
	if class and class ~= current_class then
		leave_class()
	end
	if delayed_space then
		out:write(delayed_space)
		delayed_space = nil
	end
	if class and class ~= current_class and class ~= 0 then
		enter_class(class)
	end
	out:write(html)
end

local function add_char(c, class)
	if special_chars[c] then
		add_html(special_chars[c], class)
	else
		add_html(c, class)
	end
end

local function disassemble_ligature(head)
	local type = head.id
	if type == 29 then
		if bit32.btest(head.subtype, 2) then
			local c = head.components
			lig = ""
			while c do
				lig = lig .. disassemble_ligature(c)
				c = c.next
			end
			return lig
		else
			return unicode.utf8.char(head.char)
		end
		return "?"
	else
		ebook.warn("Unsupported ligature component type " .. type .. " (" .. node.type(type) .. ")")
		return ""
	end
end

local function convert_hlist(ohead, otail, head)
	local nbsp = 0
	local debug = ebook.debug
	local soft_hyphens = ebook.get_config_bool('typo', 'soft_hyphens', 0)

	while head do
		local type = head.id
		local sub = head.subtype
		local next = head.next
		head.next = nil
		if type == 0 or type == 1 then	-- hlist or vlist
			if sub == 2 then			-- box
				debug("# Nested box")
				ohead, otail = convert_hlist(ohead, otail, head.list)
				debug("# End of nested box")
			elseif sub == 3 then			-- indent (ignored)
			else
				ebook.warn("Unsupported hlist/vlist subtype " .. sub)
			end
		elseif type == 7 then		-- disc
			if soft_hyphens and
			   head.pre and
			   head.pre.id == 29 and		-- glyph
			   head.pre.char == 45 and		-- "-"
			   not head.post and
			   not head.replace then
				-- Normal hyphenation point
				debug("> #hyphen")
				add_char(unicode.utf8.char(0xad))
			else
				ohead, otail = convert_hlist(ohead, otail, head.replace)
			end
		elseif type == 8 then		-- whatsit
			if sub == 8 then			-- user-defined
				if head.user_id == raw_html_whatsit and head.type == 115 then
					debug("> html: " .. head.value)
					add_html(head.value, node.get_attribute(head, text_class_attr))
				elseif head.user_id == anchor_whatsit and head.type == 115 then
					add_html("<span id='" .. head.value .. "'></span>")
				else
					ebook.warn("Unsupported user_whatsit ID " .. (head.user_id or "?"))
				end
			elseif sub <= 2 then			-- open, close, write
				ohead, otail = list_append(ohead, otail, head)
			else
				-- ebook.warn("Unsupported whatsit subtype " .. sub)
			end
		elseif type == 9 then		-- local_par (ignored)
		elseif type == 11 then		-- math
			if sub == 0 then			-- begin math
				leave_class()
				add_html("<span class='math'>")
			else
				leave_class()
				add_html("</span>")
			end
		elseif type == 12 then		-- glue
			if sub == 13 or sub == 14 then		-- (x)spaceskip
				if nbsp > 0 then
					debug("> #NBSP")
					delayed_space = (delayed_space or "") .. unicode.utf8.char(160)
					nbsp = 0
				else
					debug("> #SPACE")
					delayed_space = " "
				end
			elseif sub == 15 then			-- parfillskip (ignored)
			elseif sub == 0 then			-- user (ignored)
			else
				ebook.warn("Unsupported glue subtype " .. sub)
			end
		elseif type == 13 then		-- kern (ignored)
		elseif type == 14 then		-- penalty
			if head.penalty == 10000 and next and next.id == 12 then
				-- This is going to be a non-breakable space
				nbsp = 1
			elseif head.penalty == -10000 then
				delayed_space = " "
			end
		elseif type == 29 then		-- glyph
			local class = node.get_attribute(head, text_class_attr)
			nbsp = 0
			if bit32.btest(sub, 2) then		-- ligature
				lig = disassemble_ligature(head)
				debug("> #LIG: " .. lig)
				if lig_table[lig] then
					lig = lig_table[lig]
				end
				for a in string.utfcharacters(lig) do
					add_char(a, class)
				end
			else					-- regular glyph
				debug("> " .. unicode.utf8.char(head.char))
				add_char(unicode.utf8.char(head.char), class)
			end
		else
			ebook.warn("Unsupported node type " .. type .. " (" .. node.type(type) .. ")")
		end
		head = next
	end
	return ohead, otail
end

local function extract_anchors(head)
	for n in node.traverse(head) do
		local type = n.id
		if type == 0 or type == 1 then	-- hlist
			extract_anchors(n.head)
		elseif type == 8 then		-- whatsit
			if n.subtype == 8 and			-- user-defined
			   n.user_id == anchor_whatsit and n.type == 115 then
				out:write("<div id='" .. n.value .. "'></div>")
			end
		end
	end
end

local pic_counter = 0

function ebook.as_pic_ref(class)
	pic_counter = pic_counter + 1
	local name = string.format("../img/%05d.%s", pic_counter, ebook.get_config('images', 'format', 'png'))
	local div = "<div class='" .. class .. "'>"
	local img = "<img src='" .. name .. "' alt='[picture]' />"
	local undiv = "</div>\n\n"
	return div .. img .. undiv
end

local function process_contrib(head)
	local n = head
	local ohead, otail
	local debug = ebook.debug

	while n do
		local type = n.id
		local sub = n.subtype
		local next = n.next
		n.next = nil
		if type == 0 then		-- hlist
			if not n.list then
				goto cont
			end
			local par_class = node.get_attribute(n.list, par_class_attr)
			local par_id = node.get_attribute(n.list, par_id_attr)
			if sub == 6 then			-- equation
				if node.get_attribute(n, as_pic_attr) then
					extract_anchors(n)
					ohead, otail = list_append(ohead, otail, n)
					ohead, otail = list_append(ohead, otail, new_vfill())
					ohead, otail = list_append(ohead, otail, new_penalty(-10000))
					out:write(ebook.as_pic_ref("picdisplay"))
					goto cont
				end
				par_class = 1
			end
			local tag = ebook.par_class_tags[par_class] or "p"
			local class = ebook.par_classes[par_class] or ""
			local start = "<" .. tag
			if par_id and ebook.par_ids[par_id] then
				start = start .. " id='" .. ebook.par_ids[par_id] .. "'"
				ebook.par_ids[par_id] = nil
			end
			if class ~= "" then
				start = start .. " class='" .. class .. "'"
			end
			start = start .. ">"
			delayed_start_of_par = start
			delayed_space = nil
			current_class = 0
			ohead, otail = convert_hlist(ohead, otail, n.list)
			leave_class()
			if not delayed_start_of_par then
				out:write("</" .. tag .. ">\n\n")
			end
		elseif type == 1 then		-- vlist
			if node.get_attribute(n, as_pic_attr) then
				extract_anchors(n.head)
				ohead, otail = list_append(ohead, otail, n.head)
				n.head = nil
			else
				ebook.warn("Unsupported contrib vlist")
			end
		elseif type == 4 then		-- mark (ignored)
		elseif type == 8 then		-- whatsit
			if sub == 8 then			-- user-defined
				if n.user_id == raw_html_whatsit and n.type == 115 then
					debug("> html: " .. n.value)
					out:write(n.value)
				elseif n.user_id == anchor_whatsit and n.type == 115 then
					out:write("<div id='" .. n.value .. "'></div>")
				else
					ebook.warn("Unsupported user_whatsit ID " .. (n.user_id or "?"))
				end
			elseif sub <= 2 then			-- open, close, write
				ohead, otail = list_append(ohead, otail, n)
			else
				-- ebook.warn("Unsupported whatsit subtype " .. sub)
			end
		elseif type == 12 then		-- glue (ignored)
		elseif type == 13 then		-- kern (ignored)
		elseif type == 14 then		-- penalty (ignored)
		else
			ebook.warn("Unsupported contrib node type " .. type .. " (" .. node.type(type) .. ")")
		end
		::cont::
		n = next
	end

	node.slide(ohead)
	return ohead
end

-- Math conversion

local function math_learn(head, class)
	local html = head.value
	head = head.next
	if head.id ~= 18 or head.next then
		tex.error("Malformed math_learn: not a single atom")
	end
	local nuc = head.nucleus
	if nuc.id ~= 23 then
		tex.error("Malformed math_learn: nucleus is not a math_char")
	end
	math_char_def(nuc.fam, nuc.char, html, class)
end

local function math_process_char(head, tail, fam, char)
	if ebook.math_chars[fam] then
		local ch = ebook.math_chars[fam][char]
		if ch then
			local class = ebook.math_classes[fam][char] or 0
			head, tail = list_append_html(head, tail, ch, class)
		else
			ebook.warn(string.format("Math char %02x%02x not known", fam, char))
			head, tail = list_append(head, tail, new_fixme())
		end
	else
		ebook.warn("Math family " .. fam .. " not known")
		head, tail = list_append(head, tail, new_fixme())
	end
	return head, tail
end

-- \prime in superscript must be pulled out
local function math_prime_hack(head, tail, n)

	local function is_prime(n)
		return n and
		       n.id == 23 and		-- math_char
		       ebook.math_chars[n.fam] and
		       ebook.math_chars[n.fam][n.char] == "'"
	end

	if n.sup then
		local sup = n.sup
		if is_prime(sup) then
			head, tail = math_process_char(head, tail, sup.fam, sup.char)
			node.free(sup)
			n.sup = nil
		elseif sup.id == 25 then	-- sub_mlist
			while sup.head and sup.head.id == 18 and	-- noad
			      sup.head.nucleus and
			      not sup.head.sup and
			      not sup.head.sub and
			      is_prime(sup.head.nucleus) do
				local h = sup.head
				local x = h.nucleus
				head, tail = math_process_char(head, tail, x.fam, x.char)
				sup.head = h.next
				node.free(h)
			end
			if not sup.head then
				node.free(sup)
				n.sup = nil
			end
		end
	end

	return head, tail
end

-- Math styles are numbered: 0=display, 1=text, 2=script, 3=scriptscript
-- Cramping of styles is not observed.

-- What is the style of num/den of fraction typeset in style s?
local function style_to_frac(s)
	if s <= 2 then
		return s + 1
	else
		return 3
	end
end

-- What is the style of super/sub-script of an atom typeset in style s?
local function style_to_script(s)
	if s <= 1 then
		return 2
	else
		return 3
	end
end

-- 1=thin, 2=med, 3=thick, <0 if ignored in script styles
local math_spacing_table = {
	--	ord	op	bin	rel	open	close	punct	inner
	[0]={		[1]=1,	[4]=-2,	[5]=-3,				[9]=-1	},	-- ord
	[1]={	[0]=1,	[1]=1,		[5]=-3,				[9]=-1	},	-- op
	[4]={	[0]=-2,	[1]=-2,			[6]=-2,			[9]=-2	},	-- bin
	[5]={	[0]=-3,	[1]=-3,			[6]=-3,			[9]=-3	},	-- rel
	[6]={									},	-- open
	[7]={		[1]=1,	[4]=-2,	[5]=-3,				[9]=-1	},	-- close
	[8]={	[0]=-1,	[1]=-1,	[4]=-1,		[6]=-1,	[7]=-1,	[8]=-1,	[9]=-1	},	-- punct
	[9]={	[0]=-1,	[1]=1,	[4]=-2,	[5]=-3,	[6]=-1,		[8]=-1,	[9]=-1	},	-- inner
}

local math_spaces = { [0]=0, [1]=3, [2]=4, [3]=5 }

local function math_convert_list(ohead, otail, n, style)

	local debug = ebook.debug

	local function append_html(html, class)
		ohead, otail = list_append_html(ohead, otail, html, class)
	end

	local function subnode(n, style)
		local repl_id = node.get_attribute(n, math_repl_attr)
		if repl_id then				-- Node with a replacement
			local repl = ebook.math_repls[repl_id]
			append_html(repl.html, repl.class)
			return
		end

		local type = n.id
		if type == 23 then			-- math_char
			ohead, otail = math_process_char(ohead, otail, n.fam, n.char)
		elseif type == 24 then			-- sub_box
			ohead, otail = list_append(ohead, otail, n.head)
		elseif type == 25 then			-- sub_mlist
			ohead, otail = math_convert_list(ohead, otail, n.head, style)
		elseif type == 27 then			-- delim
			-- FIXME: bit 7 of small_char contains garbage, looks like LuaTeX bug
			ohead, otail = math_process_char(ohead, otail, n.small_fam, bit32.band(n.small_char, 127))
		else
			ebook.warn("Math conversion: unsupported subnode type " .. type .. " (" .. node.type(type) .. ")")
			ohead, otail = list_append(ohead, otail, new_fixme())
		end
	end

	local function sup_sub(n, style)
		ohead, otail = math_prime_hack(ohead, otail, n)
		if n.sup and n.sub then
			-- <sup> and <sub> together do not work, use a <table> instead
			append_html("<span class='mss'><span class='msp'>", 0)
			subnode(n.sup, style_to_script(style))
			append_html("</span><span class='msb'>", 0)
			subnode(n.sub, style_to_script(style))
			append_html("</span></span>", 0)
		else
			if n.sup then
				append_html("<sup>", 0)
				subnode(n.sup, style_to_script(style))
				append_html("</sup>", 0)
			end
			if n.sub then
				append_html("<sub>", 0)
				subnode(n.sub, style_to_script(style))
				append_html("</sub>", 0)
			end
		end
	end

	local function is_trivial(n)
		return not n or				-- empty
		       n.id == 23 or			-- just math_char
		       n.id == 18 and is_trivial(n.nucleus) or			-- a trivial noad with possibly non-trivial sup/sub-script
		       n.id == 25 and not n.head.next and is_trivial(n.head)	-- a trivial sub_mlist
	end

	local function find_next_atom(n)
		while n do
			local m = n
			local repl_id = node.get_attribute(n, math_repl_attr)
			if repl_id and ebook.math_repls[repl_id].type >= 0 then
				return ebook.math_repls[repl_id].type
			end
			while m.id == 17 do		-- choice
				m = m.text
			end
			local type = m.id
			if type == 18 then		-- noad
				return m.subtype
			elseif type == 19 or		-- radical
			       type == 20 or		-- fraction
			       type == 21 then		-- accent
				return 0		-- => all behave as ord
			end
		end
		return -1
	end

	local prev_atom = -1
	local inter_atom_space = 0			-- float points
	local mu_to_pt = 10/18

	local function noad_subtype(n)
		local sub = n.subtype
		if sub == 2 or sub == 3 then			-- various kinds of op atoms
			sub = 1
		elseif sub == 4 then
			-- bin atom decays to ord in certain contexts
			local next_atom = find_next_atom(n)
			if prev_atom < 0 or			-- begin of list
			   next_atom < 0 or			-- end of list
			   prev_atom >= 1 and prev_atom <= 6 or	-- op, bin, rel, open
			   prev_atom == 8 or			-- punct
			   next_atom == 5 or			-- rel
			   next_atom == 7 or			-- close
			   next_atom == 8 then			-- punct
				sub = 0
			end
		elseif sub >= 10 then				-- inner, over, vcenter behave as ord
			sub = 0
		end
		return sub
	end

	local function between_atoms(sub, style)
		local space_code = (math_spacing_table[prev_atom] or {}) [sub] or 0
		if space_code < 0 then
			if style <= 1 then
				space_code = -space_code
			else
				space_code = 0
			end
		end
		local skip = math_spaces[space_code] * mu_to_pt
		inter_atom_space = inter_atom_space + skip
		debug(string.format("Spacing: %d -> %d (style %d): type %d (%.3f pt) => total %.3f pt",
			prev_atom, sub, style, space_code, skip, inter_atom_space))

		if inter_atom_space >= 10 then
			append_html(" ")		-- U+2003
		elseif inter_atom_space >= 2.5 then
			append_html(" ")		-- U+0020
		elseif inter_atom_space >= 1 then
			append_html(" ")		-- U+2009
		end

		inter_atom_space = 0
		prev_atom = sub
	end

	local function do_list(n, style)
		while n do
			local type = n.id
			local repl_id = node.get_attribute(n, math_repl_attr)
			if repl_id then				-- Node with a replacement
				local repl = ebook.math_repls[repl_id]
				local sub = 0
				if type == 18 then
					sub = noad_subtype(n)
				end
				if repl.type >= 0 then
					sub = repl.type
				end
				between_atoms(sub, style)
				append_html(repl.html, repl.class)
			elseif type == 8 then		-- whatsit
				local sub = n.subtype
				if sub == 8 then			-- user-defined
					if n.user_id == raw_html_whatsit and n.type == 115 or
					   n.user_id == anchor_whatsit and n.type == 115
					then
						ohead, otail = list_append(ohead, otail, node.copy(n))
					else
						ebook.warn("Math conversion: Unsupported user_whatsit ID " .. (n.user_id or "?"))
					end
				elseif sub <= 2 then			-- open, close, write
					ohead, otail = list_append(ohead, otail, node.copy(n))
				else
					-- ebook.warn("Math conversion: Unsupported whatsit subtype " .. sub)
				end
			elseif type == 12 then			-- glue
				local sub = n.subtype
				local skip = 0
				if sub == 0 then			-- user
					skip = n.width / 65536
				elseif sub == 98 then			-- \nonscript: FIXME
				elseif sub == 99 then			-- mskip
					skip = n.width / 65536 * mu_to_pt
				end
				if skip ~= 0 then
					debug(string.format("Explicit glue: %.3f pt", skip))
					inter_atom_space = inter_atom_space + skip
				end
			elseif type == 13 then			-- kern (ignored)
				local sub = n.subtype
				local skip = 0
				if sub == 1 then			-- user
					skip = n.kern / 65536
				elseif sub == 99 then			-- mkern
					skip = n.kern / 65536 * mu_to_pt
				end
				if skip ~= 0 then
					debug(string.format("Explicit kern: %.3f pt", skip))
					inter_atom_space = inter_atom_space + skip
				end
			elseif type == 14 then			-- penalty (ignored)
			elseif type == 16 then			-- style
				local s = n.style
				if s == 'display' then
					style = 0
				elseif s == 'text' then
					style = 1
				elseif s == 'script' then
					style = 2
				elseif s == 'scriptscript' then
					style = 3
				else
					ebook.warn("Unknown math style " .. s)
				end
			elseif type == 17 then			-- choice
				if style == 0 then
					do_list(n.display, style)
				elseif style == 2 then
					do_list(n.text, style)
				elseif style == 3 then
					do_list(n.script, style)
				else
					do_list(n.scriptscript, style)
				end
			elseif type == 18 then			-- noad
				local sub = n.subtype;
				between_atoms(noad_subtype(n), style)
				local cls
				if sub == 10 then			-- under
					cls = "munder"
				elseif sub == 11 then			-- over
					cls = "mover"
				elseif sub == 12 then			-- vcent
					cls = "mvcent"
				end
				if cls then
					append_html("<span class='" .. cls .. "'>", 0)
				end
				if n.nucleus then
					subnode(n.nucleus, style)
				end
				if cls then
					append_html("</span>", 0)
				end
				sup_sub(n, style)
			elseif type == 19 then			-- radical
				between_atoms(0, style)
				if n.left then
					subnode(n.left, style)
				end
				if n.nucleus then
					if is_trivial(n.nucleus) then
						subnode(n.nucleus, style)
					else
						append_html("(", 0)
						subnode(n.nucleus, style)
						append_html(")", 0)
					end
				end
				sup_sub(n, style)
			elseif type == 20 then			-- fraction
				between_atoms(9, style)
				if ebook.get_config_bool('typo', 'nice_fractions', 0) then
					if n.left then
						append_html("<span class='mfl'>", 0);
						subnode(n.left, style)
						append_html("</span>", 0);
					end
					local st = style_to_frac(style)
					append_html("<span class='mf'><span class='mft'>", 0);
					ohead, otail = math_convert_list(ohead, otail, n.num, st)
					append_html("</span>", 0)
					if n.width > 0 then
						append_html("<span class='mfline'></span>", 0)
					end
					append_html("<span class='mfb'>", 0);
					ohead, otail = math_convert_list(ohead, otail, n.denom, st)
					append_html("</span></span>", 0);
					if n.right then
						append_html("<span class='mfr'>", 0);
						subnode(n.right, style)
						append_html("</span>", 0);
					end
				else
					-- a rather crude approximation for readers with broken CSS
					if n.left then
						subnode(n.left, style)
					end
					local num_trivial = is_trivial(n.num)
					local den_trivial = is_trivial(n.denom)
					if not num_trivial then append_html("(", 0) end
					ohead, otail = math_convert_list(ohead, otail, n.num, style)
					if not num_trivial then append_html(")", 0) end
					if n.width > 0 then
						append_html("/", 0)
					else
						append_html(" ", 0)
					end
					if not den_trivial then append_html("(", 0) end
					ohead, otail = math_convert_list(ohead, otail, n.denom, style)
					if not den_trivial then append_html(")", 0) end
					if n.right then
						subnode(n.right, style)
					end
				end
			elseif type == 21 then			-- accent (FIXME: special case for \hat{c} and \hat{C})
				between_atoms(0, style)
				if n.accent and
				   n.accent.id == 23 and		-- math_char
				   n.accent.char == 94 and		-- "^"
				   not n.bot_accent and
				   n.nucleus and
				   n.nucleus.id == 23 and		-- math_char
				   n.nucleus.fam == 1 and
				   (n.nucleus.char == 67 or n.nucleus.char == 99) then	-- "C" or "c"
					local ch = "Ĉ"
					if n.nucleus.char == 99 then
						ch = "ĉ"
					end
					append_html(ch)
					sup_sub(n, style)
				else
					ebook.warn("Math conversion: unsupported accent")
				end
			elseif type == 22 then			-- fence
				between_atoms(0, style)
				subnode(n.delim, style)
			elseif type == 25 then			-- sub_mlist
				between_atoms(0, style)
				ohead, otail = math_convert_list(ohead, otail, n.head, style)
			else
				ebook.warn("Math conversion: unsupported node type " .. type .. " (" .. node.type(type) .. ")")
				ohead, otail = list_append(ohead, otail, new_fixme())
			end
			n = n.next
		end
	end

	debug("*** Start of list")
	do_list(n, style)
	debug("*** End of list")
	return ohead, otail
end

local function math_convert(head, style)
	local ohead, otail
	ohead, otail = math_convert_list(ohead, otail, head, style)
	node.slide(ohead)
	return ohead
end

-- Verbatim input

function ebook.verbinput(name)
	local fn = kpse.find_file(name, true)
	if not fn then
		tex.error("Cannot find file " .. name)
		return
	end

	ebook.debug("### Including " .. fn .. " verbatim")
	local f = io.open(fn, "r")
	if not f then
		tex.error("Cannot read file " .. fn)
		return
	end

	local body = f:read("*a")
	body = html_escape(body)
	out:write("<pre><code>")
	out:write(body)
	out:write("</code></pre>\n\n")

	f:close()
end

-- Engage all callbacks

local function line_break_hack(head, is_display)
	ebook.debug("### Line-break filter called")
	if node.get_attribute(head, as_pic_attr) then
		ebook.debug("### Line-break as-pic")
		return "?"
	end
	-- ebook.debug_nodes(head)
	-- convert(head)

	-- Just wrap the contents in a line box
	local x = node.new(0, 1)	-- hbox, line
	x.list = head
	x.dir = "TLT"

	-- ebook.debug_nodes(x)
	ebook.debug("### END")
	return x
end

local function math_hack(head, disp_type, need_penalties)
	ebook.debug("### Math filter called (" .. disp_type .. ")")
	ebook.debug_nodes(head)
	if head.id == 8 and head.subtype == 8 and head.user_id == math_learn_whatsit then
		math_learn(head, node.get_attribute(head, text_class_attr))
		return head
	end
	if node.get_attribute(head, as_pic_attr) then
		ebook.debug("### Math in as-pic mode")
		head = node.mlist_to_hlist(head, disp_type, need_penalties)
	else
		head = math_convert(head, 1)
	end
	ebook.debug("### Math filter output")
	ebook.debug_nodes(head)
	return head
end

local function build_hack(state)
	ebook.debug("### Build hack called in state " .. state)
	if finished then
		-- \end inserts special penalties and if these are not
		-- propagated to the page builder, TeX never stops.
		ebook.debug("### Finished => doing nothing")
		return
	end
	local c = tex.lists["contrib_head"]
	ebook.debug_nodes(c)
	c = process_contrib(c)
	if c then
		ebook.debug("### Filtered contributions:")
		ebook.debug_nodes(c)
	end
	tex.lists["contrib_head"] = c
	ebook.debug("### END")
end

luatexbase.add_to_callback('linebreak_filter', line_break_hack, "Hack")
luatexbase.add_to_callback('mlist_to_hlist', math_hack, "Hack")
luatexbase.add_to_callback('buildpage_filter', build_hack, "Hack")
